@@ -3,76 +3,148 @@
 #include "headers.h"
 
 double
-KL_V_xi (const gsl_vector *v_V_xi)
+KL_V_xi_e (const gsl_vector *v_V_xi_e)
 {
-  params->V_xi[*params->p] = gsl_vector_get(v_V_xi, 0);
+  params->V_xi_e[*params->p] = gsl_vector_get(v_V_xi_e, 0);
   double tmpsum = 0.0, tmp;
   double KL;
   tmpsum = loglikefunc();
-  tmp = -pow (params->V_xi[*params->p] - *params->xi, 2.0);
+  tmp = -pow (params->V_xi_e[*params->p] - *params->xi, 2.0);
   KL = fabs (tmpsum + 0.5*(tmp) / *params->psi2);
   //KL = fabs (tmpsum + (tmp) / *params->psi2);
   return KL;
 }
 
-void gr_KL_V_xi (const gsl_vector *v_V_xi, void *null, gsl_vector *df)
+double
+KL_V_xi_n (const gsl_vector *v_V_xi_n)
 {
-  params->V_xi[*params->p] = gsl_vector_get(v_V_xi, 0);
-  int i, j, p;
+  int P_n=*params->P_n;
+  params->V_xi_n[*params->i* P_n+*params->p] = gsl_vector_get(v_V_xi_n, 0);
+  double tmpsum = 0.0, tmp;
+  double KL;
+  tmpsum = loglikefunc();
+  tmp = -pow (params->V_xi_n[*params->i* P_n+*params->p] - *params->xi, 2.0);
+  KL = fabs (tmpsum + 0.5*(tmp) / *params->psi2);
+  //KL = fabs (tmpsum + (tmp) / *params->psi2);
+  return KL;
+}
+
+void gr_KL_V_xi_e (const gsl_vector *v_V_xi_e, void *null, gsl_vector *df)
+{
+  int i, j, p=*params->p, pn;
+  int P_e=*params->P_e, P_n=*params->P_n;
+  params->V_xi_e[p] = gsl_vector_get(v_V_xi_e, 0);
   double tmpsum = 0.0, cov, cov2;
   int N = *params->N;
   double KL;
   int *sample_non_edges = calloc(*params->NnonE, sizeof(int));
   for (i = 0; i < *params->NE; i++) // loop over all edges
+    {
+    cov = params->V_xi_e[p]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + p];
+    cov2= params->V_psi2_e[p]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + p]; 
+    for (pn=0;pn<P_n;pn++)
       {
-      cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
-        {
-        cov += params->V_xi[p]*params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* 
-	                            *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* 
-	                              *params->P + p];
-        }
-      tmpsum += params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* 
-                                   *params->P + *params->p]*
-                (1.0 - 1.0/(1.0 + exp (-cov + 
-		params->dists[((params->E[i*2]-1)*N + params->E[i*2+1]-1)] - 0.5 * cov2)));
+      cov += params->V_xi_n[(params->E[i*2]-1)* P_n+pn]*params->XX_n[(params->E[i*2]-1)* P_n+pn] + 
+             params->V_xi_n[(params->E[i*2+1]-1)* P_n+pn]*params->XX_n[(params->E[i*2+1]-1)* P_n+pn];
+      cov2+= params->V_psi2_n[pn]*
+            (params->XX_n[(params->E[i*2]-1)* P_n+pn]+params->XX_n[(params->E[i*2+1]-1)* P_n+pn]);
       }
+    tmpsum += params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + p]*
+              (1.0 - 1.0/(1.0 + exp (-cov + params->dists[((params->E[i*2]-1)*N + params->E[i*2+1]-1)] - 0.5 * cov2)));
+    }
   sample_permutation(*params->NnonE, sample_non_edges, params->seed);
   for (j=0;j<STRATSUB1;j++) // loop over a sample of the non-edges
+    {
+    i=sample_non_edges[j];
+    cov = params->V_xi_e[p]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + p];
+    cov2= params->V_psi2_e[p]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + p]; 
+    for (pn=0;pn<P_n;pn++)
       {
-      i=sample_non_edges[j];
-      cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
-        {
-        cov += params->V_xi[p]*params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)*
-                                    *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)*
-                                      *params->P + p];
-        }
-      tmpsum += *params->NnonE/STRATSUB1*(params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)*
-                                   *params->P + *params->p]*
-                (- 1.0/(1.0 + exp (-cov +
-                params->dists[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)] - 0.5 * cov2))));
+      cov += params->V_xi_n[(params->nonE[i*2]-1)* P_n+pn]*params->XX_n[(params->nonE[i*2]-1)* P_n+pn] +
+             params->V_xi_n[(params->nonE[i*2+1]-1)* P_n+pn]*params->XX_n[(params->nonE[i*2+1]-1)* P_n+pn];
+      cov2+= params->V_psi2_n[pn]*
+            (params->XX_n[(params->nonE[i*2]-1)* P_n+pn]+params->XX_n[(params->nonE[i*2+1]-1)* P_n+pn]);
       }
-  KL = tmpsum - 0.5*fabs(params->V_xi[*params->p] - *params->xi) / *params->psi2;
+    tmpsum += *params->NnonE/STRATSUB1*(params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + p]*
+                (- 1.0/(1.0 + exp (-cov + params->dists[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)] - 0.5 * cov2))));
+      }
+  KL = tmpsum - 0.5*fabs(params->V_xi_e[p] - *params->xi) / *params->psi2;
   free(sample_non_edges);
   gsl_vector_set(df, 0, -KL);
   return;
+}
+void gr_KL_V_xi_n (const gsl_vector *v_V_xi_n, void *null, gsl_vector *df)
+{
+  int i=*params->i, p=*params->p, P_n=*params->P_n, P_e=*params->P_e, j, pe;
+  params->V_xi_n[i*P_n+p] = gsl_vector_get(v_V_xi_n, 0);
+  double tmpsum = 0.0, cov, cov2;
+  int *sample_nodes, hsum, h, Nnon, k;
+  int N = *params->N;
+  double KL;
+  int diam=*params->diam;
+  for (j = 2+diam; j < params->hopslist[i*(CONST+diam+N)+1]+2+diam; j++) // loop over all edges
+    {
+    cov = params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] +
+          params->V_xi_n[(params->hopslist[i*(CONST+diam+N)+j]-1)*P_n+p]*
+	  params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p];
+    cov2= params->V_psi2_n[i*P_n+p]*
+         (params->XX_n[i*P_n + p]+params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p]);
+    for (pe=0;pe<P_e;pe++)
+      {
+      cov += params->V_xi_e[pe]*params->XX_e[(i*N + params->hopslist[i*(CONST+diam+N)+j]-1)* P_e + pe];  
+      cov2+= params->V_psi2_e[pe]*params->XX_e[(i*N + params->hopslist[i*(CONST+diam+N)+j]-1)* P_e + pe];
+      }
+    tmpsum += (1.0 - 1.0/(1.0 + exp (-cov + params->dists[i*N + (params->hopslist[i*(CONST+diam+N)+j]-1)] - 0.5 * cov2)));
+    }
+  hsum = 0;
+  for (h=2;h<1+diam;h++) // <1+diam because we don't sample unknown edges
+    {
+    Nnon=params->hopslist[i*(CONST+diam+N)+h];
+    if (Nnon>0)
+      {
+      sample_nodes = calloc(Nnon, sizeof(int));
+      sample_permutation(Nnon, sample_nodes, params->seed);
+      for (k=0;k<STRATSUB2;k++)  // loop over some of the non-edges
+        {
+        j=params->hopslist[i*(CONST+diam+N)+2+diam+params->hopslist[i*(CONST+diam+N)+1]+hsum+sample_nodes[k]]-1;
+        cov = params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] + params->V_xi_n[j*P_n+p]* params->XX_n[j* P_n + p];
+        cov2= params->V_psi2_n[i*P_n+p]*(params->XX_n[i* P_n + p]+params->XX_n[j* P_n + p]);
+	for (pe=0;pe<P_e;pe++)
+          {
+          cov += params->V_xi_e[pe]*params->XX_e[(i*N + j)* P_e + pe];
+          cov2+= params->V_psi2_e[pe]*params->XX_e[(i*N + j)* P_e + pe];
+          }
+        tmpsum += Nnon/STRATSUB2*(- 1.0/(1.0 + exp (-cov + params->dists[i*N + j] - 0.5 * cov2)));
+        }
+      hsum += Nnon;
+      free(sample_nodes);
+      }
+    }
+  KL = tmpsum - 0.5*fabs(params->V_xi_n[i*P_n+p] - *params->xi) / *params->psi2;
+  gsl_vector_set(df, 0, -KL);
+  return;
 }
  void
-     xi_fdf (const gsl_vector *x, 
+     xi_e_fdf (const gsl_vector *x, 
              void *null, double *f, gsl_vector *df)
      {
-       *f = KL_V_xi(x);
-       gr_KL_V_xi(x, NULL, df);
+       *f = KL_V_xi_e(x);
+       gr_KL_V_xi_e(x, NULL, df);
+       return;
+     }
+void
+     xi_n_fdf (const gsl_vector *x, 
+             void *null, double *f, gsl_vector *df)
+     {
+       *f = KL_V_xi_n(x);
+       gr_KL_V_xi_n(x, NULL, df);
        return;
      }
 
 double KL_V_z_i (const gsl_vector *v_V_z_i)
 {
   int i = *params->i, d = *params->d;
-  int D = *params->D;	//, P = *params->P;
+  int D = *params->D;	
   int g;
   double tmpsum = 0.0, tmp;
   int N = *params->N;
@@ -99,34 +171,42 @@ double KL_V_z_i (const gsl_vector *v_V_z_i)
 
 void gr_KL_V_z_i (const gsl_vector *v_V_z_i, void *null, gsl_vector *df)
 {
-  int i = *params->i, j, k, d;//= *params->d;
-  int D = *params->D;     //, P = *params->p;
+  int i = *params->i, j, k, d;
+  int D = *params->D;     
   double tmp;
   for (d=0; d<D; d++)
     params->V_z[i * D + d] = gsl_vector_get(v_V_z_i, d);
   int g, p, G = *params->G;
   int N = *params->N, h, Nnon, hsum;
   int diam=*params->diam;
+  int P_n=*params->P_n;
   double *KL = calloc(D, sizeof(double));
   double *tmpsum = calloc(D, sizeof(double));
   double cov, cov2;
   int *sample_nodes;
   for (j = 2+diam; j < params->hopslist[i*(CONST+diam+N)+1]+2+diam; j++) // loop over all edges
+    {
+    tmp = 0.0;
+    for (d=0; d<D; d++)
+      tmp += pow(params->V_z[i * D + d] - params->V_z[(params->hopslist[i*(CONST+diam+N)+j]-1) * D + d], 2.0);
+    tmp = SQRT(tmp + D *(params->V_sigma2[i] + params->V_sigma2[(params->hopslist[i*(CONST+diam+N)+j]-1)]));
+    cov=0.0; cov2=0.0;
+    for (p=0;p<P_n;p++)
       {
-      tmp = 0.0;
-        for (d=0; d<D; d++)
-          tmp += pow(params->V_z[i * D + d] - params->V_z[(params->hopslist[i*(CONST+diam+N)+j]-1) * D + d], 2.0);
-      tmp = SQRT(tmp + D *(params->V_sigma2[i] + params->V_sigma2[(params->hopslist[i*(CONST+diam+N)+j]-1)]));
-      cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
-        {
-        cov += params->V_xi[p]*params->XX[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P + p];
-        }
-      for (d=0; d<D; d++)
-          tmpsum[d] +=  (params->V_z[i * D + d] - params->V_z[(params->hopslist[i*(CONST+diam+N)+j]-1) * D + d]) *
-                        (1.0 - 1.0/(1.0+exp(-cov + tmp - 0.5*cov2)));
+      cov += params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] + 
+             params->V_xi_n[(params->hopslist[i*(CONST+diam+N)+j]-1)*P_n+p]*
+	     params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p];
+      cov2+= params->V_psi2_n[p]*(params->XX_n[i* P_n + p] + params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p]);
       }
+    for (p=0;p<*params->P_e;p++)
+      {
+      cov += params->V_xi_e[p]*params->XX_e[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P_e + p];
+      cov2+= params->V_psi2_e[p]*params->XX_e[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P_e + p];
+      }
+    for (d=0; d<D; d++)
+        tmpsum[d] +=  (params->V_z[i * D + d] - params->V_z[(params->hopslist[i*(CONST+diam+N)+j]-1) * D + d]) *
+                      (1.0 - 1.0/(1.0+exp(-cov + tmp - 0.5*cov2)));
+    }
   hsum = 0;
   for (h=2;h<1+diam;h++) // <1+diam because we don't sample unknown edges
     {
@@ -143,10 +223,15 @@ void gr_KL_V_z_i (const gsl_vector *v_V_z_i, void *null, gsl_vector *df)
           tmp += pow(params->V_z[i * D + d] - params->V_z[j * D + d], 2.0);
         tmp = SQRT(tmp + D *(params->V_sigma2[i] + params->V_sigma2[j]));
         cov=0.0; cov2=0.0;
-        for (p=0;p<*params->P;p++)
+        for (p=0;p<P_n;p++)
           {
-          cov += params->V_xi[p]*params->XX[(i*N + j)* *params->P + p];
-          cov2+= params->V_psi2[p]*params->XX[(i*N + j)* *params->P + p];
+          cov += params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] + params->V_xi_n[j*P_n+p]*params->XX_n[j* P_n + p];
+          cov2+= params->V_psi2_n[p]*(params->XX_n[i* P_n + p]+params->XX_n[j* P_n + p]);
+          }
+        for (p=0;p<*params->P_e;p++)
+          {
+          cov += params->V_xi_e[p]*params->XX_e[(i*N + j)* *params->P_e + p];
+          cov2+= params->V_psi2_e[p]*params->XX_e[(i*N + j)* *params->P_e + p];
           }
         for (d=0; d<D; d++)
             tmpsum[d] +=  Nnon/STRATSUB2*(
@@ -190,7 +275,7 @@ double
 KL_V_sigma2_i (const gsl_vector *v_V_sigma2_i)
 {
   int i = *params->i, j;
-  int D = *params->D;	//, P = *params->P;
+  int D = *params->D;	
   int g, G = *params->G;
   double tmpsum = 0.0;
   double KL;
@@ -215,11 +300,12 @@ KL_V_sigma2_i (const gsl_vector *v_V_sigma2_i)
 void gr_KL_V_sigma2_i (const gsl_vector *v_V_sigma2_i, void *null, gsl_vector *df)
 {
   int i = *params->i, j, k, d = *params->d;
-  int D = *params->D;     //, P = *params->p;
+  int D = *params->D;     
   int g, G = *params->G, hsum, h;
   double tmpsum = 0.0, tmp;
   int p, N = *params->N;
   int diam=*params->diam;
+  int P_n=*params->P_n;
   double KL;
   double cov, cov2;
   double V_sigma2_i = gsl_vector_get(v_V_sigma2_i, 0);
@@ -231,10 +317,17 @@ void gr_KL_V_sigma2_i (const gsl_vector *v_V_sigma2_i, void *null, gsl_vector *d
         for (d = 0; d < D; d++)
           tmp += pow (params->V_z[i * D + d] - params->V_z[(params->hopslist[i*(CONST+diam+N)+j]-1) * D + d], 2.0);
       cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
+      for (p=0;p<P_n;p++)
         {
-        cov += params->V_xi[p]*params->XX[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P + p];
+        cov += params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] + 
+	       params->V_xi_n[(params->hopslist[i*(CONST+diam+N)+j]-1)*P_n+p]*
+	       params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p];
+        cov2+= params->V_psi2_n[p]*(params->XX_n[i* P_n + p] + params->XX_n[(params->hopslist[i*(CONST+diam+N)+j]-1)* P_n + p]);
+        }
+      for (p=0;p<*params->P_e;p++)
+        {
+        cov += params->V_xi_e[p]*params->XX_e[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P_e + p];
+        cov2+= params->V_psi2_e[p]*params->XX_e[(i*N + (params->hopslist[i*(CONST+diam+N)+j]-1))* *params->P_e + p];
         }
         tmp = SQRT (tmp + D * (V_sigma2_i + params->V_sigma2[(params->hopslist[i*(CONST+diam+N)+j]-1)]));
         tmpsum = tmpsum - D + D / (V_sigma2_i * (1.0 + exp (-cov + tmp - 0.5 * cov2)));
@@ -254,10 +347,16 @@ void gr_KL_V_sigma2_i (const gsl_vector *v_V_sigma2_i, void *null, gsl_vector *d
         for (d = 0; d < D; d++)
           tmp += pow (params->V_z[i * D + d] - params->V_z[j * D + d], 2.0);
         cov=0.0; cov2=0.0;
-        for (p=0;p<*params->P;p++)
+        for (p=0;p<P_n;p++)
           {
-          cov += params->V_xi[p]*params->XX[(i*N + j)* *params->P + p];
-          cov2+= params->V_psi2[p]*params->XX[(i*N + j)* *params->P + p];
+          cov += params->V_xi_n[i*P_n+p]*params->XX_n[i* P_n + p] + 
+	         params->V_xi_n[j*P_n+p]*params->XX_n[j* P_n + p];
+          cov2+= params->V_psi2_n[p]*(params->XX_n[i* P_n + p] + params->XX_n[j* P_n + p]);
+          }
+        for (p=0;p<*params->P_e;p++)
+          {
+          cov += params->V_xi_e[p]*params->XX_e[(i*N + j)* *params->P_e + p];
+          cov2+= params->V_psi2_e[p]*params->XX_e[(i*N + j)* *params->P_e + p];
           }
           tmp = SQRT (tmp + D * (V_sigma2_i + params->V_sigma2[j]));
           tmpsum = tmpsum + Nnon/STRATSUB2*(D / (V_sigma2_i * (1.0 + exp (-cov + tmp - 0.5 * cov2))));
@@ -354,8 +453,7 @@ void gr_KL_V_alpha_g (const gsl_vector *v_V_alpha_g, void *null, gsl_vector *df)
 double
 KL_V_nu_g (const gsl_vector *v_V_nu_g)
 {
-  int i = *params->i;	//, j, d = *params->d;
-  //int D = *params->D, P = *params->P;
+  int i = *params->i;	
   int g, G = *params->G;
   double tmpsum = 0.0, tmp;
   int N = *params->N;
@@ -417,73 +515,137 @@ void gr_KL_V_nu_g (const gsl_vector *v_V_nu_g, void *null, gsl_vector *df)
 
 
 double
-KL_V_psi2 (const gsl_vector *v_V_psi2)
+KL_V_psi2_e (const gsl_vector *v_V_psi2_e)
 {
   int D = *params->D;
-  params->V_psi2[*params->p] = gsl_vector_get(v_V_psi2, 0);
+  params->V_psi2_e[*params->p] = gsl_vector_get(v_V_psi2_e, 0);
   double tmpsum = 0.0;
   tmpsum = loglikefunc();
   double KL = tmpsum +
-	    0.5*( D * (log (params->V_psi2[*params->p])-log(*params->psi2)) -
-	    D * params->V_psi2[*params->p] / *params->psi2);
+	    0.5*( D * (log (params->V_psi2_e[*params->p])-log(*params->psi2)) -
+	    D * params->V_psi2_e[*params->p] / *params->psi2);
+  KL = fabs (KL);
+  return KL;
+}
+double
+KL_V_psi2_n (const gsl_vector *v_V_psi2_n)
+{
+  int D = *params->D;
+  params->V_psi2_n[*params->p] = gsl_vector_get(v_V_psi2_n, 0);
+  double tmpsum = 0.0;
+  tmpsum = loglikefunc();
+  double KL = tmpsum +
+	    0.5*( D * (log (params->V_psi2_n[*params->p])-log(*params->psi2)) -
+	    D * params->V_psi2_n[*params->p] / *params->psi2);
   KL = fabs (KL);
   return KL;
 }
 
-void gr_KL_V_psi2 (const gsl_vector *v_V_psi2, void * null, gsl_vector *df)
+void gr_KL_V_psi2_e (const gsl_vector *v_V_psi2_e, void * null, gsl_vector *df)
 {
-  int j, i = *params->i, p;
-  int D = *params->D;     //, P = *params->p;
-  //int g, G = *params->G;
-  //double tmpsum = 0.0;
+  int j, i = *params->i, p=*params->p, pn;
+  int D = *params->D, P_e = *params->P_e, P_n = *params->P_n;
   double tmp1=0.0, cov, cov2;
   int N = *params->N;
   double KL;
   int *sample_non_edges = calloc(*params->NnonE, sizeof(int));
-  params->V_psi2[*params->p] = gsl_vector_get(v_V_psi2, 0);
+  params->V_psi2_e[*params->p] = gsl_vector_get(v_V_psi2_e, 0);
   KL = 0.0;
   for (i = 0; i < *params->NE; i++) // loop over all edges
       {
-      cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
+      cov = params->V_xi_e[p]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + p];
+      cov2= params->V_psi2_e[p]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + p];
+      for (pn=0;pn<P_n;pn++)
         {
-        cov += params->V_xi[p]*params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* *params->P + p];
+        cov +=params->V_xi_n[(params->E[i*2]-1)*P_n+pn]*params->XX_n[(params->E[i*2]-1)* P_n + pn] +
+              params->V_xi_n[(params->E[i*2+1]-1)*P_n+pn]*params->XX_n[(params->E[i*2+1]-1)* P_n + pn];
+        cov2+= params->V_psi2_n[pn]*(params->XX_n[(params->E[i*2]-1)* P_n + pn]+params->XX_n[(params->E[i*2+1]-1)* P_n + pn]);
         }
-      tmp1 += -0.5*params->XX[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* *params->P + *params->p]/
+      tmp1 += -0.5*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + *params->p]/
                (1.0+exp(-cov + params->dists[((params->E[i*2]-1)*N + params->E[i*2+1]-1)] - 0.5*cov2));
       }
   sample_permutation(*params->NnonE, sample_non_edges, params->seed);
   for (j=0;j<STRATSUB1;j++) // loop over a sample of the non-edges
       {
       i=sample_non_edges[j];
-       cov=0.0; cov2=0.0;
-      for (p=0;p<*params->P;p++)
+      cov = params->V_xi_e[p]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + p];
+      cov2= params->V_psi2_e[p]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + p];
+      for (pn=0;pn<P_n;pn++)
         {
-        cov += params->V_xi[p]*params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* *params->P + p];
-        cov2+= params->V_psi2[p]*params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* *params->P + p];
+        cov +=params->V_xi_n[(params->nonE[i*2]-1)*P_n+pn]*params->XX_n[(params->nonE[i*2]-1)* P_n + pn] +
+              params->V_xi_n[(params->nonE[i*2+1]-1)*P_n+pn]*params->XX_n[(params->nonE[i*2+1]-1)* P_n + pn];
+        cov2+= params->V_psi2_n[pn]*(params->XX_n[(params->nonE[i*2]-1)* P_n + pn]+params->XX_n[(params->nonE[i*2+1]-1)* P_n + pn]);
         }
-      tmp1 += *params->NnonE/STRATSUB1*(-0.5*params->XX[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* *params->P + *params->p]/
+      tmp1 += *params->NnonE/STRATSUB1*(-0.5*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + *params->p]/
                (1.0+exp(-cov + params->dists[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)] - 0.5*cov2)));
       }
-  KL = tmp1 + 0.5 * (D / params->V_psi2[*params->p] - D / *params->psi2);
+  KL = tmp1 + 0.5 * (D / params->V_psi2_e[*params->p] - D / *params->psi2);
   gsl_vector_set(df, 0, -KL);
   free(sample_non_edges);
   return;
 }
 
-
+void gr_KL_V_psi2_n (const gsl_vector *v_V_psi2_n, void * null, gsl_vector *df)
+{
+  int i, p=*params->p, j, pe;
+  int D = *params->D;     
+  double tmp1=0.0, cov, cov2;
+  int N = *params->N;
+  int P_n = *params->P_n, P_e = *params->P_e;
+  double KL;
+  int *sample_non_edges = calloc(*params->NnonE, sizeof(int));
+  params->V_psi2_n[*params->p] = gsl_vector_get(v_V_psi2_n, 0);
+  KL = 0.0;
+  for (i = 0; i < *params->NE; i++) // loop over all edges
+    {
+    cov = params->V_xi_n[(params->E[i*2]-1)*P_n+p]*params->XX_n[(params->E[i*2]-1)* P_n + p] +
+          params->V_xi_n[(params->E[i*2+1]-1)*P_n+p]*params->XX_n[(params->E[i*2+1]-1)* P_n + p];
+    cov2= params->V_psi2_n[p]*(params->XX_n[(params->E[i*2]-1)* P_n + p]+params->XX_n[(params->E[i*2+1]-1)* P_n + p]);
+    for (pe=0;pe<P_e;pe++)
+      {
+      cov += params->V_xi_e[pe]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + pe];
+      cov2+= params->V_psi2_e[pe]*params->XX_e[((params->E[i*2]-1)*N + params->E[i*2+1]-1)* P_e + pe];
+      }
+    tmp1 += -0.5/(1.0+exp(-cov + params->dists[((params->E[i*2]-1)*N + params->E[i*2+1]-1)] - 0.5*cov2));
+    }
+  sample_permutation(*params->NnonE, sample_non_edges, params->seed);
+  for (j=0;j<STRATSUB1;j++) // loop over a sample of the non-edges
+    {
+    i=sample_non_edges[j];
+    cov = params->V_xi_n[(params->nonE[i*2]-1)*P_n+p]*params->XX_n[(params->nonE[i*2]-1)* P_n + p] +
+           params->V_xi_n[(params->nonE[i*2+1]-1)*P_n+p]*params->XX_n[(params->nonE[i*2+1]-1)* P_n + p];
+    cov2= params->V_psi2_n[p]*(params->XX_n[(params->nonE[i*2]-1)* P_n + p]+params->XX_n[(params->nonE[i*2+1]-1)* P_n + p]);
+    for (pe=0;pe<P_e;pe++)
+      {
+      cov += params->V_xi_e[pe]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + pe];
+      cov2+= params->V_psi2_e[pe]*params->XX_e[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)* P_e + pe];
+      }
+    tmp1 += *params->NnonE/STRATSUB1*(-0.5/(1.0+exp(-cov + params->dists[((params->nonE[i*2]-1)*N + params->nonE[i*2+1]-1)] - 0.5*cov2)));
+    }
+  KL = tmp1 + 0.5 * (D / params->V_psi2_n[*params->p] - D / *params->psi2);
+  gsl_vector_set(df, 0, -KL);
+  free(sample_non_edges);
+  return;
+}
 
  void
-     psi2_fdf (const gsl_vector *x, 
+     psi2_e_fdf (const gsl_vector *x, 
              void *null, double *f, gsl_vector *df)
      {
-       *f = KL_V_psi2(x);
-       gr_KL_V_psi2(x, NULL, df);
+       *f = KL_V_psi2_e(x);
+       gr_KL_V_psi2_e(x, NULL, df);
        return;
      }
-
-void KL_total (int *P,
+ void
+     psi2_n_fdf (const gsl_vector *x, 
+             void *null, double *f, gsl_vector *df)
+     {
+       *f = KL_V_psi2_n(x);
+       gr_KL_V_psi2_n(x, NULL, df);
+       return;
+     }
+void KL_total (int *P_n,
+  int *P_e,
   int *D,
   int *N,
   int *NE, // #edges
@@ -498,9 +660,12 @@ void KL_total (int *P,
   int *EnonE,
   int *diam,
   int *hopslist,
-  double *XX,
-  double *V_xi,
-  double *V_psi2,
+  double *XX_n,
+  double *XX_e,
+  double *V_xi_n,
+  double *V_xi_e,
+  double *V_psi2_n,
+  double *V_psi2_e,
   double *V_z,
   double *V_sigma2,
   double *V_eta,
@@ -528,7 +693,8 @@ void KL_total (int *P,
   params->g=&g;
   params->d=&d;
   params->flag=&flag;
-  params->P=P;
+  params->P_n=P_n;
+  params->P_e=P_e;
   params->D=D;
   params->N=N;
   params->NE=NE;
@@ -543,9 +709,12 @@ void KL_total (int *P,
   params->EnonE=EnonE;
   params->diam=diam; 
   params->hopslist=hopslist; 
-  params->XX=XX; // design matrix for covariates. May also be used for sender / receiver effects, etc.
-  params->V_xi=V_xi;
-  params->V_psi2=V_psi2;
+  params->XX_n=XX_n; // design matrix for node covariates. May also be used for sender / receiver effects, etc.
+  params->XX_e=XX_e; // design matrix for edge covariates. 
+  params->V_xi_n=V_xi_n;
+  params->V_xi_e=V_xi_e;
+  params->V_psi2_n=V_psi2_n;
+  params->V_psi2_e=V_psi2_e;
   params->V_z=V_z;
   params->V_sigma2=V_sigma2;
   params->V_eta=V_eta;
@@ -603,8 +772,15 @@ void KL_total (int *P,
   for (g=0;g<*G;g++)
     *KL += lgamma(V_nu[g]) - lgamma(nu[g]) - (V_nu[g]-nu[g])*(gsl_sf_psi(V_nu[g])-gsl_sf_psi(nu[g]));
   // p5-q5
-  for (p=0;p<*P;p++)
-    *KL += 0.5*(*D*(log(V_psi2[p])-log(*psi2)) - *D*(V_psi2[p]/ *psi2) - pow(V_xi[p]-xi[p],2.0)/ *psi2 +*D);
+  if (*P_n > 0)
+    for (p=0;p<*P_n;p++)
+      {
+      *KL += 0.5*(*D*(log(V_psi2_n[p])-log(*psi2)) - *D*(V_psi2_n[p]/ *psi2));
+      for (i=0;i<*N;i++)
+        *KL += 0.5*(- pow(V_xi_n[i* *P_n+p]-xi[p],2.0)/ *psi2 +*D);
+      }
+  for (p=0;p<*P_e;p++)
+    *KL += 0.5*(*D*(log(V_psi2_e[p])-log(*psi2)) - *D*(V_psi2_e[p]/ *psi2) - pow(V_xi_e[p]-xi[p],2.0)/ *psi2 +*D);
   // p6-q6
   for (g=0;g<*G;g++)
     *KL += 0.5*(*D*(log(V_omega2[g])-log(*omega2)) - *D*(V_omega2[g]/ *omega2) - pow(V_eta[g]-0.0,2.0)/ *omega2 +*D);
